import time, json
from cmflib import cmfquery
from collections import deque, defaultdict
import pandas as pd

class UniqueQueue:
    def __init__(self):
        self.queue = deque()
        self.seen = set()
    
    def enqueue(self, value):
        if value not in self.seen:
            self.queue.append(value)
            self.seen.add(value)
    
    def dequeue(self):
        if self.queue:
            value = self.queue.popleft()
            self.seen.remove(value)
            return value
        raise IndexError("dequeue from an empty queue")
    
    def __len__(self):
        return len(self.queue)
    
    def __contains__(self, value):
        return value in self.seen


async def query_execution_lineage_d3tree(mlmd_path: str, pipeline_name: str, dict_of_exe_id, uuid):
    query = cmfquery.CmfQuery(mlmd_path)
    pipeline_id = query.get_pipeline_id(pipeline_name)
    df=dict_of_exe_id[pipeline_name]
    
    #finding execution_id by comparing Execution_uuid (d09fdb26-0e9d-11ef-944f-4bf54f5aca7f) and uuid ('Prepare_u3tr')  
    result = df[df['Execution_uuid'].str[:4] == uuid]   #result = df[id: "1","Execution_type_name", "Execution_uuid"]
    execution_id=result["id"].tolist() 
    parents_set = set()
    queue = UniqueQueue()
    df = pd.DataFrame()


    parents = query.get_one_hop_parent_executions_ids(execution_id, pipeline_id) #list of parent execution ids
    dict_parents = {}
    if parents == None:
        parents = []
    dict_parents[execution_id[0]] = list(set(parents))  # [2] = [1,2,3,4] list of parent id
    parents_set.add(execution_id[0])     #created so that we can directly find execuions using execution ids
    for i in set(parents):
        queue.enqueue(i)
        parents_set.add(i)

    while len(queue) > 0:
        exe_id = queue.dequeue()
        parents = query.get_one_hop_parent_executions_ids([exe_id], pipeline_id)
        if parents == None:
            parents = [] 
        dict_parents[exe_id] = list(set(parents))
        for i in set(parents):
            queue.enqueue(i)
            parents_set.add(i)
    
    df = query.get_executions_with_execution_ids(list(parents_set))  # for execution_id get executions(complete df with all data of executions)
    df['name_uuid'] = df['Execution_type_name'] + '_' + df['Execution_uuid'] 
    result_dict = df.set_index('id')['name_uuid'].to_dict()   # {"id" : "name_uuid"} for example {"2":"Prepare_d09fdb26-0e9d-11ef-944f-4bf54f5aca7f"}

    data_organized = topological_sort(dict_parents,result_dict) # it will use topological sort to create data from parents to child pattern
    """
    data_organized format
    [[{'id': 'Prepare_d09f', 'parents': []}],  
    [{'id': 'Featurize_fae6', 'parents': ['Prepare_d09f']}], 
    [{'id': 'Train_7fe7', 'parents': ['Featurize_fae6']}]]
    """
    return data_organized

def topological_sort(input_data,execution_id_dict):
    # Initialize in-degree of all nodes to 0
    in_degree = {node: 0 for node in input_data}
    # Initialize adjacency list
    adj_list = defaultdict(list)

    # Fill the adjacency list and in-degree dictionary
    for node, dependencies in input_data.items():
        for dep in dependencies:
            adj_list[dep].append(node)
            in_degree[node] += 1

    # Queue for nodes with in-degree 0
    zero_in_degree_queue = deque([node for node, degree in in_degree.items() if degree == 0])
    topo_sorted_nodes = []

    while zero_in_degree_queue:
        current_node = zero_in_degree_queue.popleft()
        topo_sorted_nodes.append(current_node)
        for neighbor in adj_list[current_node]:
            in_degree[neighbor] -= 1
            if in_degree[neighbor] == 0:
                zero_in_degree_queue.append(neighbor)
    # Transform sorted nodes into the required output format
    parent_dict = defaultdict(list)
    # creating list of list which contains dictionary of {"id":1,parents:"execution_name"}
    for id_val in topo_sorted_nodes:   # topo_sorted_nodes = ['1','2','3','4']
        if id_val in input_data:       # input_data = {"child_id":[parents_id]}, for example {"4":['3','7','9']}
            parents = tuple(sorted(input_data[id_val]))
            # {tuple(parents): {'id':execution_name,'parents':["exec_1","exec_2","exec_3"]}
            # append id,parents to key with same parents to get all child in same list
            parent_dict[parents].append({'id': modify_exec_name(execution_id_dict[id_val]),'parents': [modify_exec_name(execution_id_dict[parent]) for parent in input_data[id_val]]})
    output_data= list(parent_dict.values()) 
    return output_data

def modify_exec_name(exec_name_uuid):
    after_first_slash=exec_name_uuid.split('/', 1)[1]
    name='_'.join(after_first_slash.rsplit('_', 1)[:-1])# 'Test-env/Prepare_d09fdb26-0e9d-11ef-944f-4bf54f5aca7f' ---> Prepare  
    uuid=exec_name_uuid.split('_')[-1].split('-')[0][:4] # 'Test-env/Prepare_d09fdb26-0e9d-11ef-944f-4bf54f5aca7f' ---> d09f 
    return (name +"_"+uuid)   # Prepare_d09f



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

"""Main parser for the cmf cli"""
import argparse
import logging
import os
import sys

from cmflib.commands import artifact, metadata, init

from cmflib.cli import CmfParserError

COMMANDS = [artifact, metadata, init]


def _find_parser(parser, cmd_cls):
    defaults = parser._defaults  # pylint: disable=protected-access
    if not cmd_cls or cmd_cls == defaults.get("func"):
        parser.print_help()
        raise CmfParserError

    actions = parser._actions  # pylint: disable=protected-access
    for action in actions:
        if not isinstance(action.choices, dict):
            # NOTE: we are only interested in subparsers
            continue
        for subparser in action.choices.values():
            _find_parser(subparser, cmd_cls)


class CmfParser(argparse.ArgumentParser):
    """Custom parser class for cmf CLI"""

    def error(self, message, cmd_cls=None):
        _find_parser(self, cmd_cls)

    def parse_args(self, args=None, namespace=None):
        args, argv = self.parse_known_args(args, namespace)
        if argv:
            msg = "unrecognized arguments: %s"
            self.error(msg % " ".join(argv), getattr(args, "func", None))
        return args


def get_parent_parser():
    """Create instances of a parser containing common arguments shared among
    all the commands.
    """
    parent_parser = argparse.ArgumentParser(add_help=False)
    return parent_parser


def get_main_parser():
    parent_parser = get_parent_parser()

    # Main parser
    desc = "Common Metadata Framework"
    parser = CmfParser(
        prog="cmf",
        description=desc,
        parents=[parent_parser],
        formatter_class=argparse.RawTextHelpFormatter,
        add_help=False,
    )

    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.",
    )

    # Sub commands
    subparsers = parser.add_subparsers(
        # required= True,
        title="Available Commands",
        metavar="COMMAND",
        dest="cmd",
        help="Use `cmf COMMAND --help` for command-specific help.",
    )

    from .utils import fix_subparsers

    fix_subparsers(subparsers)

    for cmd in COMMANDS:
        cmd.add_parser(subparsers, parent_parser)

    return parser



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import paramiko

# this is temporary - need to remove after TripleDES warning goes away from paramiko
import warnings
warnings.filterwarnings(action='ignore', module='.*paramiko.*')

class SSHremoteArtifacts:
    def download_artifacts(
        self,
        dvc_config_op,
        host: str,
        current_directory: str,
        object_name: str,
        download_loc: str,
    ):
        remote_repo = dvc_config_op["remote.ssh-storage.url"]
        user = dvc_config_op["remote.ssh-storage.user"]
        password = dvc_config_op["remote.ssh-storage.password"]
        try:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(
                paramiko.AutoAddPolicy()
            )  # this can lead to man in the middle attack, need to find another solution
            ssh.connect(host, username=user, password=password)
            sftp = ssh.open_sftp()

            dir_path = ""
            if "/" in download_loc:
                dir_path, _ = download_loc.rsplit("/", 1)
            if dir_path != "":
                # creates subfolders needed as per artifacts folder structure
                os.makedirs(dir_path, mode=0o777, exist_ok=True)  

            response = ""

            abs_download_loc = os.path.abspath(os.path.join(current_directory, download_loc))

            """"
            if object_name ends with .dir - it is a directory.
            we download .dir object with 'temp_dir' and remove 
            this after all the files from this .dir object is downloaded.
            """
            if object_name.endswith('.dir'):
                # in case of .dir, abs_download_loc is a absolute path for a folder
                os.makedirs(abs_download_loc, mode=0o777, exist_ok=True)

                 # download .dir object
                temp_dir = f"{abs_download_loc}/temp_dir"
                response = sftp.put(object_name, temp_dir)

                with open(temp_dir, 'r') as file:
                    tracked_files = eval(file.read())

                # removing temp_dir
                if os.path.exists(temp_dir):
                    os.remove(temp_dir)

                """
                object_name =  /home/user/ssh-storage/files/md5/dd/2d792b7cf6efb02231f85c6147e403.dir
                contains the path of the .dir on the artifact repo
                we need to remove the hash of the .dir from the object_name
                which will leave us with the artifact repo path
                """
                repo_path = "/".join(object_name.split("/")[:-2])
                for file_info in tracked_files:
                    relpath = file_info['relpath']
                    md5_val = file_info['md5']
                    # download_loc =  /home/user/datatslice/example-get-started/test/artifacts/raw_data
                    # md5_val = a237457aa730c396e5acdbc5a64c8453
                    # we need a2/37457aa730c396e5acdbc5a64c8453
                    formatted_md5 = md5_val[:2] + '/' + md5_val[2:]
                    temp_download_loc = f"{abs_download_loc}/{relpath}"
                    temp_object_name = f"{repo_path}/{formatted_md5}"
                    obj = sftp.put(object_name, temp_download_loc)
                    if obj:
                        print(f"object {temp_object_name} downloaded at {temp_download_loc}.")
            else:
                response = sftp.put(object_name, abs_download_loc)
            
            if response:
                stmt = f"object {object_name} downloaded at {abs_download_loc}."
                return stmt

            sftp.close()
            ssh.close()

        except TypeError as exception:
            return exception
        except Exception as exception:
            return exception



###
# Copyright (2022) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import argparse
import json
import os
from cmflib import cmfquery
from cmflib.cli.command import CmdBase


# This class export local mlmd data to a json file
class CmdMetadataExport(CmdBase):
    def run(self):

        current_directory = os.getcwd()
        full_path_to_dump = ""
        data = ""
        mlmd_data = ""

        mlmd_file_name = "./mlmd"

        # checks if mlmd filepath is given
        if self.args.file_name:
            mlmd_file_name = self.args.file_name
            current_directory = os.path.dirname(self.args.file_name)

        # checks if mlmd file is present in current directory or given directory
        if not os.path.exists(mlmd_file_name):
            return f"ERROR: {mlmd_file_name} doesn't exists in the {current_directory}."


        # setting directory where mlmd file will be dumped
        if self.args.json_file_name:
            if not os.path.isdir(self.args.json_file_name):
                temp = os.path.dirname(self.args.json_file_name)
                if temp != "":
                    current_directory = temp
                if os.path.exists(current_directory):
                    full_path_to_dump  = self.args.json_file_name
                else:
                    return f"{current_directory} doesn't exists."
            else:
                return "Provide path with file name."
        else:
            full_path_to_dump = os.getcwd() + f"/{self.args.pipeline_name}.json"

        # initialising cmfquery class
        query = cmfquery.CmfQuery(mlmd_file_name)

        # check if pipeline exists in mlmd 
        pipeline = query.get_pipeline_id(self.args.pipeline_name)

        if pipeline > 0:
            # pulling data from local mlmd file
            json_payload = query.dumptojson(self.args.pipeline_name,None)

            # write metadata into json file
            with open(full_path_to_dump, 'w') as f:
                f.write(json.dumps(json.loads(json_payload),indent=2))
                return f"SUCCESS: metadata successfully exported in {full_path_to_dump}."
        else:
            return f"{self.args.pipeline_name} doesn't exists in {mlmd_file_name}!!"


def add_parser(subparsers, parent_parser):
    PULL_HELP = "Exports local mlmd's metadata to a json file."

    parser = subparsers.add_parser(
        "export",
        parents=[parent_parser],
        description="Export local mlmd's metadata in json format to a json file.",
        help=PULL_HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    required_arguments = parser.add_argument_group("required arguments")

    required_arguments.add_argument(
        "-p",
        "--pipeline_name",
        required=True,
        help="Specify Pipeline name.",
        metavar="<pipeline_name>",
    )

    parser.add_argument(
        "-j",
        "--json_file_name",
        help="Specify json file name with full path.",
        metavar="<json_file_name>",
    )

    parser.add_argument(
        "-f", "--file_name", help="Specify mlmd file name.", metavar="<file_name>"
    )

    parser.set_defaults(func=CmdMetadataExport)
